@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2016 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2004-2019 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -572,7 +572,7 @@ class DDRProcessorImpl
 
 		if ( null == fwdSnips  ||  null == revSnips )
 			return; // error already reported
-		
+
 		try
 		{
 			DDRWriter.emit( fwdSnips, revSnips, this);
@@ -1102,9 +1102,11 @@ hunt:	for ( ExecutableElement ee : ees )
 	{
 		public String value() { return _value; }
 		public String[] defaultValue() { return _defaultValue; }
+		public String name() { return _name; }
 		
 		String _value;
 		String[] _defaultValue;
+		String _name;
 		
 		public void setValue( Object o, boolean explicit, Element e)
 		{
@@ -1117,6 +1119,20 @@ hunt:	for ( ExecutableElement ee : ees )
 			if ( explicit )
 				_defaultValue = avToArray( o, String.class);
 		}
+
+		public void setName( Object o, boolean explicit, Element e)
+		{
+			if ( ! explicit )
+				return;
+
+			_name = (String)o;
+			if ( _name.startsWith( "\"")
+				&& ! Lexicals.ISO_DELIMITED_IDENTIFIER.matcher( _name).matches()
+				|| ! Lexicals.ISO_AND_PG_REGULAR_IDENTIFIER.matcher( _name)
+					.matches()
+				)
+				msg( Kind.WARNING, e, "malformed parameter name: %s", _name);
+		}
 	}
 
 	class SQLActionsImpl extends AbstractAnnotationImpl	implements SQLActions
@@ -1262,7 +1278,7 @@ hunt:	for ( ExecutableElement ee : ees )
 			refOld = ! "".equals( _tableOld);
 			refNew = ! "".equals( _tableNew);
 
-			if ( ( refOld || refNew ) )
+			if ( refOld || refNew )
 			{
 				if ( ! Called.AFTER.equals( _called) )
 					msg( Kind.ERROR, func.func, origin,
@@ -1371,7 +1387,7 @@ hunt:	for ( ExecutableElement ee : ees )
 						sb.append("DEFERRABLE INITIALLY DEFERRED");
 						break;
 				}
-			}            
+			}
             if ( refOld || refNew )
 			{
 				sb.append( "\n\tREFERENCING");
@@ -1462,6 +1478,8 @@ hunt:	for ( ExecutableElement ee : ees )
 		boolean setof = false;
 		TypeMirror setofComponent = null;
 		boolean trigger = false;
+		TypeMirror returnTypeMapKey = null;
+		SQLType[] paramTypeAnnotations;
 
 		FunctionImpl(ExecutableElement e)
 		{
@@ -1516,6 +1534,7 @@ hunt:	for ( ExecutableElement ee : ees )
 
 			ExecutableType et = (ExecutableType)func.asType();
 			List<? extends TypeMirror> ptms = et.getParameterTypes();
+			List<? extends TypeMirror> typeArgs;
 			int arity = ptms.size();
 
 			if ( ! "".equals( type())
@@ -1533,34 +1552,16 @@ hunt:	for ( ExecutableElement ee : ees )
 					return false;
 				}
 			}
-			else if ( typu.isAssignable( typu.erasure( ret), TY_ITERATOR) )
+			else if ( null != (typeArgs = specialization( ret, TY_ITERATOR)) )
 			{
 				setof = true;
-				List<TypeMirror> pending = new LinkedList<TypeMirror>();
-				pending.add( ret);
-				while ( ! pending.isEmpty() )
+				if ( 1 != typeArgs.size() )
 				{
-					TypeMirror tm = pending.remove( 0);
-					if ( typu.isSameType( typu.erasure( tm), TY_ITERATOR) )
-					{
-						DeclaredType dt = (DeclaredType)tm;
-						List<? extends TypeMirror> typeArgs =
-							dt.getTypeArguments();
-						if ( 1 != typeArgs.size() )
-						{
-							msg( Kind.ERROR, func,
-								"Need one type argument for Iterator " +
-								"return type");
-							return false;
-						}
-						setofComponent = typeArgs.get( 0);
-						break;
-					}
-					else
-					{
-						pending.addAll( typu.directSupertypes( tm));
-					}
+					msg( Kind.ERROR, func,
+						"Need one type argument for Iterator return type");
+					return false;
 				}
+				setofComponent = typeArgs.get( 0);
 				if ( null == setofComponent )
 				{
 					msg( Kind.ERROR, func,
@@ -1583,6 +1584,8 @@ hunt:	for ( ExecutableElement ee : ees )
 					trigger = true;
 				}
 			}
+
+			returnTypeMapKey = ret;
 				
 			if ( ! setof && -1 != rows() )
 				msg( Kind.ERROR, func,
@@ -1593,6 +1596,8 @@ hunt:	for ( ExecutableElement ee : ees )
 					"a function with triggers needs void return and " +
 					"one TriggerData parameter");
 
+			collectParameterTypeAnnotations();
+
 			/*
 			 * Report any unmappable types now that could appear in
 			 * deployStrings (return type or parameter types) ... so that the
@@ -1609,6 +1614,34 @@ hunt:	for ( ExecutableElement ee : ees )
 			return true;
 		}
 
+		/*
+		 * Factored out of characterize() so it could be called if needed by
+		 * BaseUDTFunctionImpl.characterize(), which does not need anything else
+		 * from its super.characterize(). But for now it doesn't need this
+		 * either; it knows what parameters the base UDT functions take, and it
+		 * takes no heed of @SQLType annotations. Perhaps it should warn if such
+		 * annotations are used, but that's for another day.
+		 */
+		void collectParameterTypeAnnotations()
+		{
+			List<? extends VariableElement> ves = func.getParameters();
+			paramTypeAnnotations = new SQLType [ ves.size() ];
+			int i = 0;
+			for ( VariableElement ve : ves )
+			{
+				for ( AnnotationMirror am : elmu.getAllAnnotationMirrors( ve) )
+				{
+					if ( am.getAnnotationType().asElement().equals(AN_SQLTYPE) )
+					{
+						SQLTypeImpl sti = new SQLTypeImpl();
+						populateAnnotationImpl( sti, ve, am);
+						paramTypeAnnotations[i] = sti;
+					}
+				}
+				++ i;
+			}
+		}
+
 		/**
 		 * Append SQL syntax for the function's name (schema-qualified if
 		 * appropriate) and parameters, either with any defaults indicated
@@ -1637,13 +1670,24 @@ hunt:	for ( ExecutableElement ee : ees )
 				if ( complexViaInOut )
 					tms = tms.subList( 0, tms.size() - 1);
 				int s = tms.size();
+				int i = 0;
 				for ( TypeMirror tm : tms )
 				{
 					VariableElement ve = ves.next();
-					sb.append( "\n\t").append( ve.getSimpleName().toString());
+					/*
+					 * paramTypeAnnotations can be null if
+					 * collectParameterTypeAnnotations hasn't been called.
+					 * That's the case in BaseUDTFunctionImpl, but it also
+					 * overrides this method, so it isn't a problem.
+					 */
+					SQLType st = paramTypeAnnotations[i];
+					String name = null == st ? null : st.name();
+					if ( null == name )
+						name = ve.getSimpleName().toString();
+					sb.append( "\n\t").append( name);
 					sb.append( ' ');
-					sb.append( tmpr.getSQLType( tm, ve, true, dflts));
-					if ( 0 < -- s )
+					sb.append( tmpr.getSQLType( tm, ve, st, true, dflts));
+					if ( ++ i < s )
 						sb.append( ',');
 				}
 			}
@@ -1652,7 +1696,7 @@ hunt:	for ( ExecutableElement ee : ees )
 		void appendAS( StringBuilder sb)
 		{
 			if ( ! ( complexViaInOut || setof || trigger ) )
-				sb.append( func.getReturnType()).append( '=');
+				sb.append( typu.erasure( func.getReturnType())).append( '=');
 			Element e = func.getEnclosingElement();
 			if ( ! e.getKind().equals( ElementKind.CLASS) )
 				msg( Kind.ERROR, func,
@@ -1682,7 +1726,7 @@ hunt:	for ( ExecutableElement ee : ees )
 				else if ( setof )
 					sb.append( "pg_catalog.RECORD");
 				else
-					sb.append( tmpr.getSQLType( func.getReturnType(), func));
+					sb.append( tmpr.getSQLType( returnTypeMapKey, func));
 			}
 			sb.append( "\n\tLANGUAGE ");
 			if ( Trust.SANDBOXED.equals( trust()) )
@@ -1741,6 +1785,39 @@ hunt:	for ( ExecutableElement ee : ees )
 			rslt [ rslt.length - 1 ] = sb.toString();
 			return rslt;
 		}
+
+		/**
+		 * Test whether the type {@code tm} is, directly or indirectly,
+		 * a specialization of generic type {@code dt}.
+		 * @param tm a type to be checked
+		 * @param dt known generic type to check for
+		 * @return null if {@code tm} does not extend {@code dt}, otherwise the
+		 * list of type arguments with which it specializes {@code dt}
+		 */
+		List<? extends TypeMirror> specialization(
+			TypeMirror tm, DeclaredType dt)
+		{
+			if ( ! typu.isAssignable( typu.erasure( tm), dt) )
+				return null;
+
+			List<TypeMirror> pending = new LinkedList<TypeMirror>();
+			pending.add( tm);
+			while ( ! pending.isEmpty() )
+			{
+				tm = pending.remove( 0);
+				if ( typu.isSameType( typu.erasure( tm), dt) )
+					return ((DeclaredType)tm).getTypeArguments();
+				pending.addAll( typu.directSupertypes( tm));
+			}
+			/*
+			 * This is a can't-happen: tm is assignable to dt but has no
+			 * supertype that's dt? Could throw an AssertionError, but returning
+			 * an empty list will lead the caller to report an error, and that
+			 * will give more information about the location in the source being
+			 * compiled.
+			 */
+			return Collections.emptyList();
+		}
 	}
 
 	static enum BaseUDTFunctionID
@@ -2494,7 +2571,7 @@ hunt:	for ( ExecutableElement ee : ees )
 		 */
 		String getSQLType(TypeMirror tm, Element e)
 		{
-			return getSQLType( tm, e, false, false);
+			return getSQLType( tm, e, null, false, false);
 		}
 
 
@@ -2506,6 +2583,8 @@ hunt:	for ( ExecutableElement ee : ees )
 		 * @param tm Represents the type whose corresponding SQL type is wanted.
 		 * @param e Annotated element (chiefly for use as a location hint in
 		 * diagnostic messages).
+		 * @param st {@code SQLType} annotation, or null if none, explicitly
+		 * given for the element.
 		 * @param contravariant Indicates that the element whose type is wanted
 		 * is a function parameter and should be given the widest type that can
 		 * be assigned to it. If false, find the narrowest type that a function
@@ -2513,7 +2592,7 @@ hunt:	for ( ExecutableElement ee : ees )
 		 * @param withDefault Indicates whether any specified default value
 		 * information should also be included in the "type" string returned.
 		 */
-		String getSQLType(TypeMirror tm, Element e,
+		String getSQLType(TypeMirror tm, Element e, SQLType st,
 			boolean contravariant, boolean withDefault)
 		{
 			boolean array = false;
@@ -2522,15 +2601,10 @@ hunt:	for ( ExecutableElement ee : ees )
 			
 			String[] defaults = null;
 			
-			for ( AnnotationMirror am : elmu.getAllAnnotationMirrors( e) )
+			if ( null != st )
 			{
-				if ( am.getAnnotationType().asElement().equals( AN_SQLTYPE) )
-				{
-					SQLTypeImpl sti = new SQLTypeImpl();
-					populateAnnotationImpl( sti, e, am);
-					rslt = sti.value();
-					defaults = sti.defaultValue();
-				}
+				rslt = st.value();
+				defaults = st.defaultValue();
 			}
 
 			if ( tm.getKind().equals( TypeKind.ARRAY) )
@@ -2676,7 +2750,7 @@ hunt:	for ( ExecutableElement ee : ees )
 
 	// expression intended to match SQL types that are arrays
 	static final Pattern arrayish =
-		Pattern.compile( "(?si:(?:\\[\\s*\\d*\\s*\\]|ARRAY)\\s*)$");
+		Pattern.compile( "(?si:(?:\\[\\s*+\\d*+\\s*+\\]|ARRAY)\\s*+)$");
 
 	/**
 	 * Work around bizarre javac behavior that silently supplies an Error
