@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004-2018 Tada AB and other contributors, as listed below.
+ * Copyright (c) 2004-2019 Tada AB and other contributors, as listed below.
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the The BSD 3-Clause License
@@ -12,8 +12,9 @@
  *   PostgreSQL Global Development Group
  *   Chapman Flack
  */
+#include <time.h>
+
 #include <postgres.h>
-#include <utils/nabstime.h>
 #include <utils/datetime.h>
 
 #include "pljava/Backend.h"
@@ -458,7 +459,7 @@ static int32 Timestamp_getTimeZone_dd(double dt)
 
 int32 Timestamp_getCurrentTimeZone(void)
 {
-	return Timestamp_getTimeZone((pg_time_t)GetCurrentAbsoluteTime());
+	return Timestamp_getTimeZone((pg_time_t)time(NULL));
 }
 
 extern void Timestamp_initialize(void);
