@@ -1,8 +1,14 @@
 /*
- * Copyright (c) 2004, 2005, 2006 TADA AB - Taby Sweden
- * Distributed under the terms shown in the file COPYRIGHT
- * found in the root folder of this project or at
- * http://eng.tada.se/osprojects/COPYRIGHT.html
+ * Copyright (c) 2004-2019 Tada AB and other contributors, as listed below.
+ *
+ * All rights reserved. This program and the accompanying materials
+ * are made available under the terms of the The BSD 3-Clause License
+ * which accompanies this distribution, and is available at
+ * http://opensource.org/licenses/BSD-3-Clause
+ *
+ * Contributors:
+ *   Tada AB
+ *   Chapman Flack
  *
  * @author Thomas Hallgren
  */
@@ -18,7 +24,7 @@ extern "C" {
 #include <commands/trigger.h>
 
 /**********************************************************************
- * The TriggerData java class extends the NativeStruct and provides JNI
+ * The TriggerData java class provides JNI
  * access to some of the attributes of the TriggerData structure.
  * 
  * @author Thomas Hallgren
@@ -27,7 +33,7 @@ extern "C" {
 /*
  * Create the org.postgresql.pljava.TriggerData object.
  */
-extern jobject TriggerData_create(TriggerData* triggerData);
+extern jobject pljava_TriggerData_create(TriggerData* triggerData);
 
 /*
  * Obtains the returned Tuple after trigger has been processed.
@@ -35,7 +41,9 @@ extern jobject TriggerData_create(TriggerData* triggerData);
  * is connected (and that a longer-lived memory context than SPI's is selected,
  * if the caller wants the result to survive SPI_finish).
  */
-extern HeapTuple TriggerData_getTriggerReturnTuple(jobject jtd, bool* wasNull);
+extern HeapTuple pljava_TriggerData_getTriggerReturnTuple(
+	jobject jtd, bool* wasNull);
+extern void pljava_TriggerData_initialize(void);
 
 #ifdef __cplusplus
 }
